@@ -26,7 +26,7 @@
 
 extern Window get_window(Window w, Atom prop);
 extern Source<bool> disabled;
-extern Source<Window> current_window;
+extern Source<Window> current_app_window;
 extern Source<ActionListDiff *> stroke_app;
 extern bool in_proximity;
 
@@ -35,6 +35,8 @@ Grabber *grabber = 0;
 static unsigned int ignore_mods[4] = { 0, LockMask, Mod2Mask, LockMask | Mod2Mask };
 static unsigned char device_mask_data[2];
 static XIEventMask device_mask;
+static unsigned char raw_mask_data[3];
+static XIEventMask raw_mask;
 
 template <class X1, class X2> class BiMap {
 	std::map<X1, X2> map1;
@@ -218,24 +220,23 @@ void Grabber::unminimize() {
 	activate(w, CurrentTime);
 }
 
-const char *Grabber::state_name[3] = { "None", "Button", "Select" };
+const char *Grabber::state_name[4] = { "None", "Button", "Select", "Raw" };
 
 Grabber::Grabber() : children(ROOT) {
 	current = BUTTON;
 	suspended = 0;
 	suspend();
-	disabled = false;
 	active = true;
 	grabbed = NONE;
 	xi_grabbed = false;
-	xi_devs_grabbed = false;
+	xi_devs_grabbed = GrabNo;
 	grabbed_button.button = 0;
 	grabbed_button.state = 0;
 	cursor_select = XCreateFontCursor(dpy, XC_crosshair);
 	init_xi();
 	prefs.excluded_devices.connect(new IdleNotifier(sigc::mem_fun(*this, &Grabber::update)));
 	prefs.button.connect(new IdleNotifier(sigc::mem_fun(*this, &Grabber::update)));
-	current_class = fun2(&get_wm_class, current_window, stroke_app);
+	current_class = fun2(&get_wm_class, current_app_window, stroke_app);
 	current_class->connect(new IdleNotifier(sigc::mem_fun(*this, &Grabber::update)));
 	disabled.connect(new IdleNotifier(sigc::mem_fun(*this, &Grabber::set)));
 	update();
@@ -288,6 +289,14 @@ bool Grabber::init_xi() {
 	XISetMask(device_mask.mask, XI_ButtonRelease);
 	XISetMask(device_mask.mask, XI_Motion);
 
+	raw_mask.deviceid = XIAllDevices;
+	raw_mask.mask = raw_mask_data;
+	raw_mask.mask_len = sizeof(raw_mask_data);
+	memset(raw_mask.mask, 0, raw_mask.mask_len);
+	XISetMask(raw_mask.mask, XI_ButtonPress);
+	XISetMask(raw_mask.mask, XI_ButtonRelease);
+	XISetMask(raw_mask.mask, XI_RawMotion);
+
 	XIEventMask global_mask;
 	unsigned char data[2] = { 0, 0 };
 	global_mask.deviceid = XIAllDevices;
@@ -364,7 +373,7 @@ void Grabber::new_device(XIDeviceInfo *info) {
 		}
 }
 
-Grabber::XiDevice::XiDevice(Grabber *parent, XIDeviceInfo *info) : supports_pressure(false), absolute(false), num_buttons(0) {
+Grabber::XiDevice::XiDevice(Grabber *parent, XIDeviceInfo *info) : supports_pressure(false), absolute(false), scale_x(1.0), scale_y(1.0), num_buttons(0) {
 	dev = info->deviceid;
 	name = info->name;
 	master = info->attachment;
@@ -375,8 +384,14 @@ Grabber::XiDevice::XiDevice(Grabber *parent, XIDeviceInfo *info) : supports_pres
 			num_buttons = b->num_buttons;
 		} else if (dev_class->type == ValuatorClass) {
 			XIValuatorClassInfo *v = (XIValuatorClassInfo*)dev_class;
-			if ((v->number == 0 || v->number == 1) && v->mode != XIModeRelative)
+			if ((v->number == 0 || v->number == 1) && v->mode != XIModeRelative) {
 				absolute = true;
+				if (v-> number == 0)
+					scale_x = (double)DisplayWidth(dpy, DefaultScreen(dpy)) / (double)(v->max - v->min);
+				else
+					scale_y = (double)DisplayHeight(dpy, DefaultScreen(dpy)) / (double)(v->max - v->min);
+
+			}
 			if (v->number == 2) {
 				pressure_min = v->min;
 				pressure_max = v->max;
@@ -428,18 +443,19 @@ void Grabber::grab_xi(bool grab) {
 				i->second->grab_button(*j, grab);
 }
 
-void Grabber::XiDevice::grab_device(bool grab) {
-	if (!grab) {
+void Grabber::XiDevice::grab_device(GrabState grab) {
+	if (grab == GrabNo) {
 		XIUngrabDevice(dpy, dev, CurrentTime);
 		if (current_dev && current_dev->dev == dev)
 			xinput_pressed.clear();
 		return;
 	}
-	XIGrabDevice(dpy, dev, ROOT, CurrentTime, None, GrabModeAsync, GrabModeAsync, False, &device_mask);
+	XIGrabDevice(dpy, dev, ROOT, CurrentTime, None, GrabModeAsync, GrabModeAsync, False,
+			grab == GrabYes ? &device_mask : &raw_mask);
 }
 
-void Grabber::grab_xi_devs(bool grab) {
-	if (!xi_devs_grabbed == !grab)
+void Grabber::grab_xi_devs(GrabState grab) {
+	if (xi_devs_grabbed == grab)
 		return;
 	xi_devs_grabbed = grab;
 	for (DeviceMap::iterator i = xi_devs.begin(); i != xi_devs.end(); ++i)
@@ -449,7 +465,14 @@ void Grabber::grab_xi_devs(bool grab) {
 void Grabber::set() {
 	bool act = !suspended && ((active && !disabled.get()) || (current != NONE && current != BUTTON));
 	grab_xi(act && current != SELECT);
-	grab_xi_devs(act && current == NONE);
+	if (!act)
+		grab_xi_devs(GrabNo);
+	else if (current == NONE)
+		grab_xi_devs(GrabYes);
+	else if (current == RAW)
+		grab_xi_devs(GrabRaw);
+	else
+		grab_xi_devs(GrabNo);
 	State old = grabbed;
 	grabbed = act ? current : NONE;
 	if (old == grabbed)
@@ -568,27 +591,22 @@ Window find_wm_state(Window w) {
 	return found;
 }
 
-// sets w to 0 if the window is a frame
-Window get_app_window(Window &w) {
+Window get_app_window(Window w) {
 	if (!w)
 		return w;
-	if (frame_win.contains1(w)) {
-		Window w2 = frame_win.find1(w);
-		w = 0;
-		return w2;
-	}
-	if (frame_child.contains1(w)) {
-		Window w2 = frame_child.find1(w);
-		if (w != w2)
-			w = 0;
-		return w2;
-	}
+
+	if (frame_win.contains1(w))
+		return frame_win.find1(w);
+
+	if (frame_child.contains1(w))
+		return frame_child.find1(w);
+
 	Window w2 = find_wm_state(w);
 	if (w2) {
 		frame_child.add(w, w2);
 		if (w2 != w) {
 			w = w2;
-			XSelectInput(dpy, w2, EnterWindowMask | LeaveWindowMask | StructureNotifyMask | PropertyChangeMask);
+			XSelectInput(dpy, w2, StructureNotifyMask | PropertyChangeMask);
 		}
 		return w2;
 	}
