@@ -45,8 +45,9 @@ extern Source<bool> disabled;
 
 bool experimental = false;
 int verbosity = 0;
-const char *versions[] = { "-0.4.1", "-0.4.0", "", NULL };
-Source<Window> current_window(None);
+const char *prefs_versions[] = { "-0.5.5", "-0.4.1", "-0.4.0", "", NULL };
+const char *actions_versions[] = { "-0.4.1", "-0.4.0", "", NULL };
+Source<Window> current_app_window(None);
 std::string config_dir;
 Win *win = NULL;
 Display *dpy;
@@ -57,11 +58,12 @@ Grabber::XiDevice *current_dev = 0;
 std::set<guint> xinput_pressed; // TODO get rid of
 
 static bool show_gui = false;
+static bool no_dbus = false;
 
 static int argc;
 static char **argv;
 
-static Window current_app = 0, ping_window = 0;
+static Window ping_window = 0;
 static boost::shared_ptr<Trace> trace;
 static std::map<guint, guint> core_inv_map;
 
@@ -133,6 +135,7 @@ public:
 	static bool idle() { return !handler->child; }
 
 	virtual void motion(RTriple e) {}
+	virtual void raw_motion(RTriple e, bool, bool) {}
 	virtual void press(guint b, RTriple e) {}
 	virtual void release(guint b, RTriple e) {}
 	virtual void press_master(guint b, Time t) {}
@@ -241,6 +244,40 @@ public:
 	virtual Grabber::State grab_mode() { return Grabber::NONE; }
 };
 
+class ButtonHandler : public Handler {
+	RModifiers mods;
+	guint button, real_button;
+public:
+	ButtonHandler(RModifiers mods_, guint button_) : mods(mods_), button(button_), real_button(0) {}
+	virtual void press(guint b, RTriple e) {
+		if (current_dev->master) {
+			if (!real_button)
+				real_button = b;
+			if (real_button == b)
+				b = button;
+			XTestFakeMotionEvent(dpy, DefaultScreen(dpy), e->x, e->y, 0);
+			XTestFakeButtonEvent(dpy, b, true, CurrentTime);
+		}
+	}
+	virtual void motion(RTriple e) {
+		if (current_dev->master)
+			XTestFakeMotionEvent(dpy, DefaultScreen(dpy), e->x, e->y, 0);
+	}
+	// TODO: Handle Proximity
+	virtual void release(guint b, RTriple e) {
+		if (current_dev->master) {
+			if (real_button == b)
+				b = button;
+			XTestFakeMotionEvent(dpy, DefaultScreen(dpy), e->x, e->y, 0);
+			XTestFakeButtonEvent(dpy, b, false, CurrentTime);
+		}
+		if (!xinput_pressed.size())
+			parent->replace_child(NULL);
+	}
+	virtual std::string name() { return "Button"; }
+	virtual Grabber::State grab_mode() { return Grabber::NONE; }
+};
+
 static void bail_out();
 
 static void bail_out() {
@@ -333,13 +370,22 @@ static void update_core_mapping() {
 static inline float abs(float x) { return x > 0 ? x : -x; }
 
 class AbstractScrollHandler : public Handler {
+	bool have_x, have_y;
 	float last_x, last_y;
 	Time last_t;
 	float offset_x, offset_y;
 	Glib::ustring str;
+	int orig_x, orig_y;
 
 protected:
-	AbstractScrollHandler() : last_t(0), offset_x(0.0), offset_y(0.0) {}
+	AbstractScrollHandler() : last_t(0), offset_x(0.0), offset_y(0.0) {
+		if (!prefs.move_back.get() || current_dev->absolute)
+			return;
+		Window dummy1, dummy2;
+		int dummy3, dummy4;
+		unsigned int dummy5;
+		XQueryPointer(dpy, ROOT, &dummy1, &dummy2, &orig_x, &orig_y, &dummy3, &dummy4, &dummy5);
+	}
 	virtual void fake_wheel(int b1, int n1, int b2, int n2) {
 		for (int i = 0; i<n1; i++)
 			fake_click(b1);
@@ -349,22 +395,41 @@ protected:
 	static float curve(float v) {
 		return v * exp(log(abs(v))/3);
 	}
+protected:
+	void move_back() {
+		if (!prefs.move_back.get() || current_dev->absolute)
+			return;
+		XTestFakeMotionEvent(dpy, DefaultScreen(dpy), orig_x, orig_y, 0);
+	}
 public:
-	virtual void motion(RTriple e) {
-		if (!last_t || abs(e->x-last_x) > 100 || abs(e->y-last_y) > 100) {
+	virtual void raw_motion(RTriple e, bool abs_x, bool abs_y) {
+		float dx = abs_x ? (have_x ? e->x - last_x : 0) : e->x;
+		float dy = abs_y ? (have_y ? e->y - last_y : 0) : e->y;
+
+		if (abs_x) {
 			last_x = e->x;
+			have_x = true;
+		}
+
+		if (abs_y) {
 			last_y = e->y;
+			have_y = true;
+		}
+
+		if (!last_t) {
 			last_t = e->t;
 			return;
 		}
+
 		if (e->t == last_t)
 			return;
-		double factor = (prefs.scroll_invert.get() ? 1.0 : -1.0) * prefs.scroll_speed.get();
-		offset_x += factor * curve((e->x-last_x)/(e->t-last_t))*(e->t-last_t)/20.0;
-		offset_y += factor * curve((e->y-last_y)/(e->t-last_t))*(e->t-last_t)/10.0;
-		last_x = e->x;
-		last_y = e->y;
+
+		int dt = e->t - last_t;
 		last_t = e->t;
+
+		double factor = (prefs.scroll_invert.get() ? 1.0 : -1.0) * prefs.scroll_speed.get();
+		offset_x += factor * curve(dx/dt)*dt/20.0;
+		offset_y += factor * curve(dy/dt)*dt/10.0;
 		int b1 = 0, n1 = 0, b2 = 0, n2 = 0;
 		if (abs(offset_x) > 1.0) {
 			n1 = (int)floor(abs(offset_x));
@@ -396,27 +461,31 @@ public:
 class ScrollHandler : public AbstractScrollHandler {
 	RModifiers mods;
 	std::map<guint, guint> map;
+	int orig_x, orig_y;
 public:
 	ScrollHandler(RModifiers mods_) : mods(mods_) {
 		// If you want to use buttons > 9, you're on your own..
 		map[1] = 0; map[2] = 0; map[3] = 0; map[8] = 0; map[9] = 0;
 	}
-	virtual void motion(RTriple e) {
+	virtual void raw_motion(RTriple e, bool abs_x, bool abs_y) {
 		if (xinput_pressed.size())
-			AbstractScrollHandler::motion(e);
+			AbstractScrollHandler::raw_motion(e, abs_x, abs_y);
 	}
 	virtual void press_master(guint b, Time t) {
 		fake_core_button(b, false);
 	}
 	virtual void release(guint b, RTriple e) {
-		if (!in_proximity && !xinput_pressed.size())
-			parent->replace_child(0);
+		if (in_proximity || xinput_pressed.size())
+			return;
+		parent->replace_child(0);
+		move_back();
 	}
 	virtual void proximity_out() {
 		parent->replace_child(0);
+		move_back();
 	}
 	virtual std::string name() { return "Scroll"; }
-	virtual Grabber::State grab_mode() { return Grabber::NONE; }
+	virtual Grabber::State grab_mode() { return Grabber::RAW; }
 };
 
 class ScrollAdvancedHandler : public AbstractScrollHandler {
@@ -428,21 +497,20 @@ public:
 		AbstractScrollHandler::fake_wheel(b1, n1, b2, n2);
 		rb = 0;
 	}
-	virtual void motion(RTriple e) {
-		AbstractScrollHandler::motion(e);
-	}
 	virtual void release(guint b, RTriple e) {
 		Handler *p = parent;
 		p->replace_child(NULL);
 		p->release(b, e);
+		move_back();
 	}
 	virtual void press(guint b, RTriple e) {
 		Handler *p = parent;
 		p->replace_child(NULL);
 		p->press(b, e);
+		move_back();
 	}
 	virtual std::string name() { return "ScrollAdvanced"; }
-	virtual Grabber::State grab_mode() { return Grabber::NONE; }
+	virtual Grabber::State grab_mode() { return Grabber::RAW; }
 };
 
 // Hack so that we don't have to move stuff around so much
@@ -669,14 +737,19 @@ static void icccm_client_message(Window w, Atom a, Time t) {
 }
 
 static void activate_window(Window w, Time t) {
+	static XAtom _NET_ACTIVE_WINDOW("_NET_ACTIVE_WINDOW");
 	static XAtom _NET_WM_WINDOW_TYPE("_NET_WM_WINDOW_TYPE");
 	static XAtom _NET_WM_WINDOW_TYPE_DOCK("_NET_WM_WINDOW_TYPE_DOCK");
 	static XAtom WM_PROTOCOLS("WM_PROTOCOLS");
 	static XAtom WM_TAKE_FOCUS("WM_TAKE_FOCUS");
 
+	if (w == get_window(ROOT, *_NET_ACTIVE_WINDOW))
+		return;
+
 	Atom window_type = get_atom(w, *_NET_WM_WINDOW_TYPE);
 	if (window_type == *_NET_WM_WINDOW_TYPE_DOCK)
 		return;
+
 	XWMHints *wm_hints = XGetWMHints(dpy, w);
 	if (wm_hints) {
 		bool input = wm_hints->input;
@@ -684,14 +757,18 @@ static void activate_window(Window w, Time t) {
 		if (!input)
 			return;
 	}
+
+	if (!has_atom(w, *WM_PROTOCOLS, *WM_TAKE_FOCUS))
+		return;
+
+	XWindowAttributes attr;
+	if (XGetWindowAttributes(dpy, w, &attr) && attr.override_redirect)
+		return;
+
 	if (verbosity >= 3)
 		printf("Giving focus to window 0x%lx\n", w);
 
-	bool take_focus = has_atom(w, *WM_PROTOCOLS, *WM_TAKE_FOCUS);
-	if (take_focus)
-		icccm_client_message(w, *WM_TAKE_FOCUS, t);
-	else
-		XSetInputFocus(dpy, w, RevertToParent, t);
+	icccm_client_message(w, *WM_TAKE_FOCUS, t);
 }
 
 void Trace::start(Trace::Point p) {
@@ -709,6 +786,36 @@ void Trace::end() {
 	end_();
 }
 
+static void get_timeouts(TimeoutType type, int *init, int *final) {
+	switch (type) {
+		case TimeoutOff:
+			*init = 0;
+			*final = 0;
+			break;
+		case TimeoutConservative:
+			*init = 750;
+			*final = 750;
+			break;
+		case TimeoutDefault:
+			*init = 250;
+			*final = 250;
+			break;
+		case TimeoutMedium:
+			*init = 100;
+			*final = 100;
+			break;
+		case TimeoutAggressive:
+			*init = 50;
+			*final = 75;
+			break;
+		case TimeoutFlick:
+			*init = 30;
+			*final = 50;
+			break;
+		default:;
+	}
+}
+
 class StrokeHandler : public Handler, public sigc::trackable {
 	guint button;
 	RPreStroke cur;
@@ -837,9 +944,10 @@ protected:
 			return parent->replace_child(NULL);
 		}
 		RModifiers mods = act->prepare();
-		if (IS_CLICK(act)) {
+		if (IS_CLICK(act))
 			act = Button::create((Gdk::ModifierType)0, b);
-		}
+		else IF_BUTTON(act, b)
+			return parent->replace_child(new ButtonHandler(mods, b));
 		if (IS_IGNORE(act))
 			return parent->replace_child(new IgnoreHandler(mods));
 		if (IS_SCROLL(act))
@@ -862,8 +970,15 @@ protected:
 	}
 public:
 	StrokeHandler(guint b, RTriple e) : button(b), is_gesture(false), drawing(false), last(e), orig(e),
-	use_timeout(prefs.init_timeout.get()),
-	init_timeout(prefs.init_timeout.get()), final_timeout(prefs.final_timeout.get()), radius(16) {}
+	init_timeout(prefs.init_timeout.get()), final_timeout(prefs.final_timeout.get()), radius(16) {
+		const std::map<std::string, TimeoutType> &dt = prefs.device_timeout.ref();
+		std::map<std::string, TimeoutType>::const_iterator j = dt.find(current_dev->name);
+		if (j != dt.end())
+			get_timeouts(j->second, &init_timeout, &final_timeout);
+		else
+			get_timeouts(prefs.timeout_profile.get(), &init_timeout, &final_timeout);
+		use_timeout = init_timeout;
+	}
 	virtual void init() {
 		if (grabber->is_instant(button))
 			return do_instant();
@@ -894,8 +1009,8 @@ protected:
 		update_core_mapping();
 	}
 	virtual void press(guint b, RTriple e) {
-		if (current_app)
-			activate_window(current_app, e->t);
+		if (current_app_window.get())
+			activate_window(current_app_window.get(), e->t);
 		replace_child(new StrokeHandler(b, e));
 	}
 public:
@@ -988,6 +1103,7 @@ public:
 	void handle_enter_leave(XEvent &ev);
 	void handle_event(XEvent &ev);
 	void handle_xi2_event(XIDeviceEvent *event);
+	void handle_raw_motion(XIRawEvent *event);
 	void report_xi2_event(XIDeviceEvent *event, const char *type);
 	~Main();
 };
@@ -1059,7 +1175,7 @@ Main::Main() : kit(0) {
 		printf(_("Couldn't open display.\n"));
 		exit(EXIT_FAILURE);
 	}
-	if (start_dbus() < 0) {
+	if (!no_dbus && start_dbus() < 0) {
 		printf(_("Easystroke is already running, showing configuration window instead.\n"));
 		send_dbus("");
 		exit(EXIT_SUCCESS);
@@ -1121,9 +1237,10 @@ void Main::usage(char *me, bool good) {
 	printf("Options:\n");
 	printf("  -c, --config-dir       Directory for config files\n");
 	printf("      --display          X Server to contact\n");
-	printf("  -x  --no-xi            Don't use the Xinput extension\n");
+	printf("  -D  --no-dbus          Don't try to register as a DBus service\n");
 	printf("  -e  --experimental     Start in experimental mode\n");
 	printf("  -g, --show-gui         Show the configuration dialog on startup\n");
+	printf("  -x  --disable          Start disabled\n");
 	printf("  -v, --verbose          Increase verbosity level\n");
 	printf("  -h, --help             Display this help and exit\n");
 	printf("      --version          Output version information and exit\n");
@@ -1152,6 +1269,8 @@ std::string Main::parse_args_and_init_gtk() {
 		{"experimental",0,0,'e'},
 		{"show-gui",0,0,'g'},
 		{"verbose",0,0,'v'},
+		{"no-dbus",0,0,'D'},
+		{"disabled",0,0,'x'},
 		{0,0,0,0}
 	};
 	std::string display;
@@ -1179,7 +1298,7 @@ std::string Main::parse_args_and_init_gtk() {
 	oldHandler = XSetErrorHandler(xErrorHandler);
 	oldIOHandler = XSetIOErrorHandler(xIOErrorHandler);
 
-	while ((opt = getopt_long(argc, argv, "c:egv", long_opts2, 0)) != -1) {
+	while ((opt = getopt_long(argc, argv, "c:egvDx", long_opts2, 0)) != -1) {
 		switch (opt) {
 			case 'c':
 				config_dir = optarg;
@@ -1190,6 +1309,12 @@ std::string Main::parse_args_and_init_gtk() {
 			case 'v':
 				verbosity++;
 				break;
+			case 'D':
+				no_dbus = true;
+				break;
+			case 'x':
+				disabled.set(true);
+				break;
 			case 'd':
 			case 'n':
 			case 'g':
@@ -1207,7 +1332,10 @@ void Main::create_config_dir() {
 		config_dir += "/.easystroke";
 	}
 	struct stat st;
-	if (lstat(config_dir.c_str(), &st) == -1) {
+	char *name = canonicalize_file_name(config_dir.c_str());
+
+	// check if the directory does not exist
+	if (lstat(name, &st) == -1) {
 		if (mkdir(config_dir.c_str(), 0777) == -1) {
 			printf(_("Error: Couldn't create configuration directory \"%s\"\n"), config_dir.c_str());
 			exit(EXIT_FAILURE);
@@ -1217,11 +1345,14 @@ void Main::create_config_dir() {
 			printf(_("Error: \"%s\" is not a directory\n"), config_dir.c_str());
 			exit(EXIT_FAILURE);
 		}
+
+
 	}
+	free (name);
 	config_dir += "/";
 }
 
-extern Window get_app_window(Window &w);
+extern Window get_app_window(Window w);
 
 void Main::handle_enter_leave(XEvent &ev) {
 	if (ev.xcrossing.mode == NotifyGrab)
@@ -1230,16 +1361,9 @@ void Main::handle_enter_leave(XEvent &ev) {
 		return;
 	Window w = ev.xcrossing.window;
 	if (ev.type == EnterNotify) {
-		current_window.set(w);
-		current_app = get_app_window(w);
+		current_app_window.set(get_app_window(w));
 		if (verbosity >= 3)
-			printf("Entered window 0x%lx -> 0x%lx\n", w, current_app);
-	} else if (ev.type == LeaveNotify) {
-		if (ev.xcrossing.window != current_window.get())
-			return;
-		if (verbosity >= 3)
-			printf("Left window 0x%lx\n", w);
-		current_window.set(None);
+			printf("Entered window 0x%lx -> 0x%lx\n", w, current_app_window.get());
 	} else printf("Error: Bogus Enter/Leave event\n");
 }
 
@@ -1253,8 +1377,8 @@ void Main::handle_event(XEvent &ev) {
 		return;
 
 	case PropertyNotify:
-		if (current_window.get() == ev.xproperty.window && ev.xproperty.atom == XA_WM_CLASS)
-			current_window.notify();
+		if (current_app_window.get() == ev.xproperty.window && ev.xproperty.atom == XA_WM_CLASS)
+			current_app_window.notify();
 		return;
 
 	case ButtonPress:
@@ -1286,15 +1410,10 @@ void Main::handle_event(XEvent &ev) {
 	}
 }
 
-void Main::report_xi2_event(XIDeviceEvent *event, const char *type) {
-	printf("%s (XI2): ", type);
-	if (event->detail)
-		printf("%d ", event->detail);
-	printf("(%.3f, %.3f) - (", event->root_x, event->root_y);
-
+static void print_coordinates(XIValuatorState *valuators, double *values) {
 	int n = 0;
-	for (int i = event->valuators.mask_len - 1; i >= 0; i--)
-		if (XIMaskIsSet(event->valuators.mask, i)) {
+	for (int i = valuators->mask_len - 1; i >= 0; i--)
+		if (XIMaskIsSet(valuators->mask, i)) {
 			n = i+1;
 			break;
 		}
@@ -1305,12 +1424,19 @@ void Main::report_xi2_event(XIDeviceEvent *event, const char *type) {
 			first = false;
 		else
 			printf(", ");
-		if (XIMaskIsSet(event->valuators.mask, i))
-			printf("%.3f", event->valuators.values[elt++]);
+		if (XIMaskIsSet(valuators->mask, i))
+			printf("%.3f", values[elt++]);
 		else
 			printf("*");
-
 	}
+}
+
+void Main::report_xi2_event(XIDeviceEvent *event, const char *type) {
+	printf("%s (XI2): ", type);
+	if (event->detail)
+		printf("%d ", event->detail);
+	printf("(%.3f, %.3f) - (", event->root_x, event->root_y);
+	print_coordinates(&event->valuators, event->valuators.values);
 	printf(") at t = %ld\n", event->time);
 }
 
@@ -1358,10 +1484,41 @@ void Main::handle_xi2_event(XIDeviceEvent *event) {
 			}
 			H->motion(create_triple(event->root_x, event->root_y, event->time));
 			break;
+		case XI_RawMotion:
+			handle_raw_motion((XIRawEvent *)event);
+			break;
 		case XI_HierarchyChanged:
 			grabber->hierarchy_changed((XIHierarchyEvent *)event);
 	}
 }
+
+void Main::handle_raw_motion(XIRawEvent *event) {
+	if (!current_dev || current_dev->dev != event->deviceid)
+		return;
+	double x = 0.0, y = 0.0;
+	bool abs_x = current_dev->absolute;
+	bool abs_y = current_dev->absolute;
+	int i = 0;
+
+	if (XIMaskIsSet(event->valuators.mask, 0))
+		x = event->raw_values[i++];
+	else
+		abs_x = false;
+
+	if (XIMaskIsSet(event->valuators.mask, 1))
+		y = event->raw_values[i++];
+	else
+		abs_y = false;
+
+	if (verbosity >= 5) {
+		printf("Raw motion (XI2): (");
+		print_coordinates(&event->valuators, event->raw_values);
+		printf(") at t = %ld\n", event->time);
+	}
+
+	H->raw_motion(create_triple(x * current_dev->scale_x, y * current_dev->scale_y, event->time), abs_x, abs_y);
+}
+
 #undef H
 
 bool Main::handle(Glib::IOCondition) {
